import imp
from fvcore.common.registry import Registry
import torch.nn as nn
import torch
from models import resnet
from collections import OrderedDict
import os
from pathlib import Path

MODEL_REGISTRY = Registry('MODEL')
MODEL_REGISTRY.__doc__ = """
Registry for video model.
The registered object will be called with `obj(**vars)`.
The call should return a `torch.nn.Module` object.
"""

def build_model(args, test_mode: bool=False) -> (nn.Module, str):
    """
    Args:
        args: all options defined in opts.py and num_classes
        test_mode:
    Returns:
        network model
        architecture name
    """
    #model = MODEL_REGISTRY.get(args.backbone_net)(**vars(args))
    if args.alg == "fixmatch":
        model = MODEL_REGISTRY.get(args.backbone_net)(pretrained=args.imagenet_pretrained)
        model.fc = nn.Linear(model.fc.in_features, args.num_classes)
    elif args.alg == "paws":
        model = resnet.__dict__[args.backbone_net]()
        hidden_dim = 2048
        output_dim = 128
        model.fc = nn.Linear(output_dim, args.num_classes)
        model.encoder = torch.nn.Sequential(OrderedDict([
            ('fc1', torch.nn.Linear(hidden_dim, hidden_dim)),
            ('bn1', torch.nn.BatchNorm1d(hidden_dim)),
            ('relu1', torch.nn.ReLU(inplace=True)),
            ('fc2', torch.nn.Linear(hidden_dim, hidden_dim)),
            ('bn2', torch.nn.BatchNorm1d(hidden_dim)),
            ('relu2', torch.nn.ReLU(inplace=True)),
            ('fc3', torch.nn.Linear(hidden_dim, output_dim))
        ]))
        # model.fc should be initialized already

    # TODO:
    network_name = model.network_name if hasattr(model, 'network_name') else args.backbone_net
    
    SSL = "SL" if args.ulb_loss_ratio == 0 else "FM"
    arch_name = "{dataset}/{type}_{arch_name}".format(
        dataset=args.dataset, type = SSL, arch_name=network_name)
    if args.nlabels is not None:
        arch_name += f'/nlabs{str(args.nlabels)}'
    if args.prefix != '':
        arch_name += f'/{args.prefix}'
    # add setting info only in training
    if not test_mode:
        arch_name += "-{}{}-bs{}-e{}".format(args.lr_scheduler, "-syncbn" if args.sync_bn else "",
                                             args.batch_size, args.epochs)
    folder = os.path.join(args.logdir, arch_name)
    if not os.path.exists(folder):
        run_id = "0"
    else:
        run_id = str(len([f for f in os.listdir(folder) if os.path.isdir(os.path.join(folder, f))]))
    arch_name = os.path.join(arch_name, run_id)
    
    if args.evaluate and args.pretrained is not None:
        # add arch_name
        arch_name = Path(args.pretrained).parent.absolute()
        arch_name = os.path.relpath(arch_name, args.logdir) # get relative path
    return model, arch_name


if __name__ == "__main__":
    @MODEL_REGISTRY.register()
    def abc():
        return 0

    print(MODEL_REGISTRY._obj_map.keys())
