from ast import arg
from distutils.command.config import config
import os
from os.path import join
import builtins
import shutil
import time
import platform
import sys
import random
import warnings

import wandb

import torch
import torch.nn as nn
import torch.nn.parallel
import torch.backends.cudnn as cudnn
import torch.distributed as dist
import torchvision.transforms as transforms
import torch.optim
import torch.utils.data
import torch.utils.data.distributed
import torch.multiprocessing as mp
from torch.optim import lr_scheduler
from torch.utils.tensorboard import SummaryWriter
from fvcore.nn import flop_count, parameter_count
from torchvision.datasets import ImageFolder

# from laplace import Laplace

import numpy as np

from utils.fixmatch_utils import fixMatchDataset
from utils.data_utils import data_subselection, combine_dataset, load_pseudo, add_pseudo_data

from models import build_model
from utils.utils import (train, validate, evaluate, build_dataflow, get_augmentor,
                         save_checkpoint, get_class_distribution, unfreeze_encoder,
                         get_mean_and_std, setup_settings_in_yaml, safe_load_image, freeze_encoder)

from utils.lr_scheduler import CosineWarmupScheduler
from opts import arg_parser

warnings.filterwarnings("ignore", category=UserWarning)

if os.environ.get('WSC', None):
    os.system("taskset -p 0xfffffffffffffffffffffffffffffffffffffffffff %d > /dev/null 2>&1" % os.getpid())

def main():
    global args
    parser = arg_parser()
    args = parser.parse_args()
    args = setup_settings_in_yaml(args)
    if args.gpu_id is not None:
        os.environ['CUDA_VISIBLE_DEVICES'] = args.gpu_id
    if args.seed is not None:
        random.seed(args.seed)
        torch.manual_seed(args.seed)
        np.random.seed(args.seed)
        cudnn.deterministic = True
    
    if args.hostfile != '':
        curr_node_name = platform.node().split(".")[0]
        with open(args.hostfile) as f:
            nodes = [x.strip() for x in f.readlines() if x.strip() != '']
            master_node = nodes[0].split(" ")[0]
        for idx, node in enumerate(nodes):
            if curr_node_name in node:
                args.rank = idx
                break
        args.world_size = len(nodes)
        args.dist_url = "tcp://{}:10598".format(master_node)

    args.distributed = args.world_size > 1 or args.multiprocessing_distributed
    ngpus_per_node = torch.cuda.device_count()
    # print(ngpus_per_node)
    if args.multiprocessing_distributed:
        # Since we have ngpus_per_node processes per node, the total world_size
        # needs to be adjusted accordingly
        args.world_size = ngpus_per_node * args.world_size
        # Use torch.multiprocessing.spawn to launch distributed processes: the
        # main_worker process function
        mp.spawn(main_worker, nprocs=ngpus_per_node, args=(ngpus_per_node, args))
    else:
        # Simply call main_worker function
        main_worker(args.gpu, ngpus_per_node, args)


def main_worker(gpu, ngpus_per_node, args):
    cudnn.benchmark = args.cudnn_benchmark
    args.gpu = gpu

    # num_classes, train_list_name, val_list_name, test_list_name, filename_seperator, image_tmpl, filter_video, label_file = get_dataset_config(args.dataset, args.loader_type)
    # args.num_classes = num_classes

    if args.gpu is not None:
        print(f"{platform.node().split('.')[0]}, Use GPU: {args.gpu} for training")

    if args.distributed:
        if args.dist_url == "env://" and args.rank == -1:
            args.rank = int(os.environ["RANK"])
        if args.multiprocessing_distributed:
            # For multiprocessing distributed training, rank needs to be the
            # global rank among all the processes
            args.rank = args.rank * ngpus_per_node + gpu
        dist.init_process_group(backend=args.dist_backend, init_method=args.dist_url,
                                world_size=args.world_size, rank=args.rank)
    # suppress printing if not master
    if args.multiprocessing_distributed and args.rank != 0:
        def print_pass(*args, **kwargs):
            pass
        builtins.print = print_pass

    args.input_channels = 3

    model, arch_name = build_model(args)

    mean, std = get_mean_and_std(model, args)

    model = model.cuda(args.gpu)
    model.eval()

    dummy_data_shape = (1, args.input_channels, args.input_size, args.input_size)
    dummy_data = torch.rand(dummy_data_shape).cuda(args.gpu)

    if args.rank == 0:
        try:
            flops = flop_count(model, (dummy_data, ))[0]['conv']
        except:
            flops = 0
        params = parameter_count(model)['']
        torch.cuda.empty_cache()

    if args.distributed:
        # For multiprocessing distributed, DistributedDataParallel constructor
        # should always set the single device scope, otherwise,
        # DistributedDataParallel will use all available devices.
        if args.gpu is not None:
            torch.cuda.set_device(args.gpu)
            # When using a single GPU per process and per
            # DistributedDataParallel, we need to divide the batch size
            # ourselves based on the total number of GPUs we have
            # the batch size should be divided by number of nodes as well
            args.batch_size = int(args.batch_size / args.world_size)
            args.workers = int(args.workers / ngpus_per_node)

            if args.sync_bn:
                process_group = torch.distributed.new_group(list(range(args.world_size)))
                model = torch.nn.SyncBatchNorm.convert_sync_batchnorm(model, process_group)
            model.cuda(args.gpu)
            model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[args.gpu])
        else:
            model.cuda()
            # DistributedDataParallel will divide and allocate batch_size to all
            # available GPUs if device_ids are not set
            model = torch.nn.parallel.DistributedDataParallel(model)
    elif args.gpu is not None:
        torch.cuda.set_device(args.gpu)
        model = model.cuda(args.gpu)
    else:
        # DataParallel will divide and allocate batch_size to all available GPUs
        # assign rank to 0
        model = torch.nn.DataParallel(model).cuda()
        args.rank = 0

    if args.pretrained is not None:
        print("=> using pre-trained model '{}'".format(args.pretrained))
        if args.gpu is None:
            checkpoint = torch.load(args.pretrained, map_location='cpu')
        else:
            checkpoint = torch.load(args.pretrained, map_location='cuda:{}'.format(args.gpu))
        if args.transfer:
            new_dict = {}
            for k, v in checkpoint['state_dict'].items():
                # TODO: a better approach:
                if k.replace("module.", "").startswith("fc"):
                    continue
                new_dict[k] = v
        else:
            new_dict = checkpoint['state_dict']
        msg = model.load_state_dict(new_dict, strict=False)
        print(msg)
        del checkpoint  # dereference seems crucial
        torch.cuda.empty_cache()
    else:
        print("=> creating model '{}'".format(arch_name))

    # define loss function (criterion) and optimizer
    train_criterion = nn.CrossEntropyLoss().cuda(args.gpu)
    val_criterion = nn.CrossEntropyLoss().cuda(args.gpu)

    # Data loading code
    val_augmentor = get_augmentor(False, args.input_size, mean=mean, std=std)

    # val_dataset = ImageFolder(os.path.join(args.datadir, args.val_folder),
    #                           transform=val_augmentor, loader=safe_load_image)
    if type(args.val_folder) == list:
        val_dataset = []
        for folder in args.val_folder:
            val_dataset.append(fixMatchDataset(os.path.join(args.datadir, folder),
                                transform=val_augmentor, loader=safe_load_image))
        val_dataset = combine_dataset(val_dataset)
    else:
        val_dataset = fixMatchDataset(os.path.join(args.datadir, args.val_folder),
                                transform=val_augmentor, loader=safe_load_image)
    if args.patch_eval is not None:
        lb_idx = data_subselection(val_dataset.samples, args, patch_list = args.patch_eval, flag = "testing")
        val_dataset.samples = [val_dataset.samples[idx] for idx in lb_idx]
        val_dataset.targets = [val_dataset.targets[idx] for idx in lb_idx]
        val_dataset.imgs = [val_dataset.imgs[idx] for idx in lb_idx]
    val_dataset.regenerate_data()
    
    print("Validation dataset: ", np.unique(val_dataset.targets, return_counts = True))
    val_loader = build_dataflow(val_dataset, is_train=False, batch_size=args.batch_size, workers=args.workers, is_distributed=args.distributed)

    log_folder = os.path.join(args.logdir, arch_name)
    if args.abs_log_dir is not None:
        log_folder = args.abs_log_dir
    if args.rank == 0:
        if not os.path.exists(log_folder):
            os.makedirs(log_folder)

    weighted_sampler = None
            

    # load unlabeled dataset
    unlabel_augmentor = get_augmentor(True, args.input_size, mean=mean, std=std)
    if type(args.unlabeled_folder) == list:
        unlabel_dataset = []
        for folder in args.unlabeled_folder:
            unlabel_dataset.append(fixMatchDataset(os.path.join(args.datadir, folder), 
                                    transform = unlabel_augmentor, loader=safe_load_image, if_strong = True))
        unlabel_dataset = combine_dataset(unlabel_dataset)
    else:
        unlabel_dataset = fixMatchDataset(os.path.join(args.datadir, args.unlabeled_folder), 
                                    transform = unlabel_augmentor, loader=safe_load_image, if_strong = True)
    print("Unlabel dataset: ", np.unique(unlabel_dataset.targets, return_counts = True))
    
    unlabel_dataset.regenerate_data()
    unlabel_loader = build_dataflow(unlabel_dataset, is_train=True, batch_size=args.batch_size * args.uratio,
                                    workers=args.workers, is_distributed=args.distributed, sampler=weighted_sampler)


    train_augmentor = get_augmentor(True, args.input_size, mean=mean, std=std)
    if type(args.train_folder) == list:
        train_dataset = []
        for folder in args.train_folder:
            train_dataset.append(fixMatchDataset(os.path.join(args.datadir, folder),
                                transform=train_augmentor, loader=safe_load_image))
        train_dataset = combine_dataset(train_dataset)
    else:
        train_dataset = fixMatchDataset(os.path.join(args.datadir, args.train_folder),
                                transform=train_augmentor, loader=safe_load_image)
    
    patch_list = args.patch_train if args.patch_train is not None else None
    lb_idx = data_subselection(train_dataset.samples, args, patch_list = patch_list, flag = "training")
    train_dataset.samples = [train_dataset.samples[idx] for idx in lb_idx]
    train_dataset.targets = [train_dataset.targets[idx] for idx in lb_idx]
    train_dataset.imgs = [train_dataset.imgs[idx] for idx in lb_idx]
    weighted_sampler = None
    if args.pseudo_label is not None:
        # add pseudo labels from unlabeled data
        print('loading pseudo labels')
        names, plabels, weight = load_pseudo(args.pseudo_label)
        train_dataset, counts, props, weights = add_pseudo_data(unlabel_dataset, train_dataset, names, plabels, weight)
        class_weights_all = torch.tensor([1.0]*len(train_dataset.targets))
        n_pesudo = len(counts)
        class_weights_all[-n_pesudo:] = torch.tensor(weights)
        weighted_sampler = torch.utils.data.WeightedRandomSampler(weights=class_weights_all,num_samples=len(class_weights_all),replacement=True)
        # weighted_sampler = torch.utils.data.WeightedRandomSampler(weights=class_weights_all,num_samples=len(class_weights_all),replacement=True)
    if args.evaluate:
        is_train = False
    else:
        is_train = True
    train_dataset.regenerate_data() # regenerate train data
    train_loader = build_dataflow(train_dataset, is_train=is_train, batch_size=args.batch_size, workers=args.workers, is_distributed=args.distributed, sampler=weighted_sampler)
    print("Train dataset: ", np.unique(train_dataset.targets, return_counts = True))
    if args.evaluate:
        #val_top1, val_top5, val_losses, val_speed = validate(val_loader, model, val_criterion, gpu_id=args.gpu, rank=args.rank)
        if args.rank == 0:
            if args.bayesian:
                model = Laplace(model, 'classification',
                    subset_of_weights='last_layer',
                    hessian_structure='kron')
                model.fit(train_loader)
                model.optimize_prior_precision(method='marglik')
            else:
                model.eval()
            logfile_val = open(os.path.join(log_folder, 'val_by_hl.txt'), 'w')
            val_acc, _, _, _ = evaluate(val_loader, model, val_criterion, gpu_id=args.gpu, rank=args.rank, logfile = logfile_val)
            logfile_val.close()
            logfile_train = open(os.path.join(log_folder, 'train_by_hl.txt'), 'w')
            train_acc, _, _, _ = evaluate(train_loader, model, val_criterion, gpu_id=args.gpu, rank=args.rank, logfile = logfile_train)
            logfile_train.close()
            print('Validation Acc: %.4f'%(val_acc))
        return

    if args.opt == "SGD":
        optimizer = torch.optim.SGD(model.parameters(), args.lr if args.warmup_lr is None else args.warmup_lr,
                                momentum=args.momentum,
                                weight_decay=args.weight_decay,
                                nesterov=args.nesterov)
    elif args.opt == "Adam":
        optimizer = torch.optim.Adam(model.parameters(), args.lr if args.warmup_lr is None else args.warmup_lr,
                                weight_decay=args.weight_decay)
    else:
        NotImplementedError
    
    # fine-tune optimizer
    if args.alg == "paws":
        unfreeze_encoder(model)
        freeze_encoder(model)
        params_to_update = []
        for name,param in model.named_parameters():
            if param.requires_grad == True:
                params_to_update.append(param)
        optimizer_fineTune = torch.optim.Adam(params_to_update, args.lr,
                                    weight_decay=args.weight_decay)
        unfreeze_encoder(model)
    else:
        optimizer_fineTune = None

    if args.lr_scheduler == 'step':
        scheduler = lr_scheduler.StepLR(optimizer, args.lr_steps[0], gamma=0.1)
    elif args.lr_scheduler == 'multisteps':
        scheduler = lr_scheduler.MultiStepLR(optimizer, args.lr_steps, gamma=0.1)
    elif args.lr_scheduler == 'cosine':
        scheduler = lr_scheduler.CosineAnnealingLR(optimizer, args.epochs, eta_min=0)
    elif args.lr_scheduler == 'plateau':
        scheduler = lr_scheduler.ReduceLROnPlateau(optimizer, 'min', verbose=True)
    elif args.lr_scheduler == 'warmupcosine':
        scheduler = lr_scheduler.LambdaLR(optimizer, CosineWarmupScheduler(args.lr, args.warmup_lr, args.epochs, args.warmup_epochs))

    best_top1 = 0.0
    if args.auto_resume:
        checkpoint_path = os.path.join(log_folder, 'checkpoint.pth.tar')
        if os.path.exists(checkpoint_path):
            args.resume = checkpoint_path
    # optionally resume from a checkpoint
    if args.resume:
        if args.rank == 0:
            logfile = open(os.path.join(log_folder, 'log.log'), 'a')
        if os.path.isfile(args.resume):
            if args.gpu is None:
                checkpoint = torch.load(args.resume, map_location='cpu')
            else:
                checkpoint = torch.load(args.resume, map_location='cuda:{}'.format(args.gpu))
            args.start_epoch = checkpoint['epoch']
            # TODO: handle distributed version
            best_top1 = checkpoint['best_top1']
            if not isinstance(best_top1, float):
                if args.gpu is not None:
                    best_top1 = best_top1.to(args.gpu)
                else:
                    best_top1 = best_top1.cuda()
            model.load_state_dict(checkpoint['state_dict'])
            optimizer.load_state_dict(checkpoint['optimizer'])
            try:
                scheduler.load_state_dict(checkpoint['scheduler'])
            except Exception as e:
                pass
            if args.rank == 0:
                print(f"=> loaded checkpoint '{args.resume}' (epoch {checkpoint['epoch']})", flush=True)
                print(f"=> loaded checkpoint '{args.resume}' (epoch {checkpoint['epoch']})", file=logfile, flush=True)
            del checkpoint  # dereference seems crucial
            torch.cuda.empty_cache()
        else:
            raise ValueError(f"Checkpoint is not found: {args.resume}")
    else:
        if os.path.exists(os.path.join(log_folder, 'log.log')) and args.rank == 0:
            shutil.copyfile(os.path.join(log_folder, 'log.log'), os.path.join(
                log_folder, 'log.log.{}'.format(int(time.time()))))
        if args.rank == 0:
            logfile = open(os.path.join(log_folder, 'log.log'), 'w')

    if args.rank == 0:
        command = " ".join(sys.argv)
        summary_writer = SummaryWriter(log_folder)
        if args.ulb_loss_ratio == 0:
            alg_name = "SL"
        else:
            alg_name = "FM"
        args.log_folder = log_folder
        
        prefix = '' if args.prefix is None else args.prefix
        group_dir = '' if args.group_dir is None else args.group_dir
            
        wandb.init(
            project = 'CryoEM_Img',
            group = args.group_dir,
            name = prefix + args.dataset + alg_name + f'_nlab{str(len(train_dataset.samples))}_ur{str(int(args.ulb_loss_ratio))}',
            config = args
        )
        print(command, flush=True)
        print(args, flush=True)
        # print(model, flush=True)
        print(command, file=logfile, flush=True)
        # print(f"FLops: {flops}, Parameters: {params}", flush=True)
        print(args, file=logfile, flush=True)

    if args.resume == '' and args.rank == 0:
        print(model, file=logfile, flush=True)
        # print(f"FLops: {flops}G, Parameters: {params}", flush=True, file=logfile)

    try:
        #lrs = list(set(scheduler.get_lr()))
        #if len(lrs) != 1:
        #    print(f"Get multiple learning rates: {lrs}")
        #lr = lrs[0]
        lr = scheduler.optimizer.param_groups[0]['lr']
    except Exception as e:
        lr = None
    for epoch in range(args.start_epoch, args.epochs):
        print(f'Epoch {epoch}: learning rate = {lr:.6f}', flush=True)
        # train for one epoch
        train_top1, train_top5, train_losses, train_unsup_losses, train_speed, speed_data_loader, train_steps = \
            train(train_loader, model, train_criterion, optimizer, epoch + 1, args,
                  unlabel_loader = unlabel_loader,
                  display=args.print_freq, label_smoothing=args.label_smoothing,
                  clip_gradient=args.clip_gradient, gpu_id=args.gpu, rank=args.rank,
                  precise_bn=args.precise_bn, optimizer_fineTune = optimizer_fineTune)
        if args.distributed:
            dist.barrier()

        eval_this_epoch = True
        if args.lazy_eval:
            if (epoch + 1) % 10 == 0 or (epoch + 1) >= args.epochs * 0.9:
                eval_this_epoch = True
            else:
                eval_this_epoch = False

        if eval_this_epoch:
            # evaluate on validation set
            val_top1, val_top5, val_losses, val_speed = validate(val_loader, model, val_criterion,
                                                                 gpu_id=args.gpu, rank=args.rank)
        else:
            val_top1, val_top5, val_losses, val_speed = 0.0, 0.0, 0.0, 0.0

        # update current learning rate
        if args.lr_scheduler == 'plateau':
            scheduler.step(val_losses)
        else:
            scheduler.step(epoch+1)

        if args.distributed:
            dist.barrier()

        # only logging at rank 0
        if args.rank == 0:
            # print(f'Train: [{epoch + 1:03d}/{args.epochs:03d}]\tLoss: {train_losses:4.4f}\t'
            #       f'Top@1: {train_top1:.4f}\tTop@5: {train_top5:.4f}\tSpeed: {train_speed * 1000.0:.2f} ms/batch\t'
            #       f'Data loading: {speed_data_loader * 1000.0:.2f} ms/batch', flush=True)
            print(f'Train: [{epoch + 1:03d}/{args.epochs:03d}]\tunSupLoss: {train_unsup_losses:4.4f}\t')
            print(f'Train: [{epoch + 1:03d}/{args.epochs:03d}]\tunSupLoss: {train_unsup_losses:4.4f}\t', file=logfile, flush=True)
            print(f'Train: [{epoch + 1:03d}/{args.epochs:03d}]\tLoss: {train_losses:4.4f}\t'
                  f'Top@1: {train_top1:.4f}\tTop@5: {train_top5:.4f}\tSpeed: {train_speed * 1000.0:.2f} ms/batch\t'
                  f'Data loading: {speed_data_loader * 1000.0:.2f} ms/batch', file=logfile, flush=True)

            print(f'Val  : [{epoch + 1:03d}/{args.epochs:03d}]\tLoss: {val_losses:4.4f}\t'
                  f'Top@1: {val_top1:.4f}\tTop@5: {val_top5:.4f}\tSpeed: {val_speed * 1000.0:.2f} ms/batch', file=logfile, flush=True)

            # remember best prec@1 and save checkpoint
            is_best = val_top1 > best_top1
            best_top1 = max(val_top1, best_top1)

            save_dict = {'epoch': epoch + 1,
                         'arch': arch_name,
                         'state_dict': model.state_dict(),
                         'best_top1': best_top1,
                         'optimizer': optimizer.state_dict(),
                         'scheduler': scheduler.state_dict()
                         }

            save_checkpoint(save_dict, is_best, filepath=log_folder)
            try:
                # get_lr get all lrs for every layer of current epoch, assume the lr for all layers are identical
                lr = scheduler.optimizer.param_groups[0]['lr']
            except Exception as e:
                lr = None

            if lr is not None:
                summary_writer.add_scalar('learning-rate', lr, epoch + 1)
            summary_writer.add_scalar('val-top1', val_top1, epoch + 1)
            summary_writer.add_scalar('val-loss', val_losses, epoch + 1)
            summary_writer.add_scalar('train-top1', train_top1, epoch + 1)
            summary_writer.add_scalar('train-loss', train_losses, epoch + 1)
            summary_writer.add_scalar('train-unsuploss', train_unsup_losses, epoch + 1)
            summary_writer.add_scalar('best-val-top1', best_top1, epoch + 1)
            wandb.log({"learning-rate": lr, "val-top1": val_top1, 'val-loss': val_losses,
                       "train-top1": train_top1, "train-loss": train_losses, "train-unsuploss": train_unsup_losses,
                       "best-val-top1": best_top1})
        if args.distributed:
            dist.barrier()
    #val_top1, val_top5, val_losses, val_speed = validate(val_loader, model, val_criterion, gpu_id=args.gpu, rank=args.rank)
    print(f"Best validation acc: {str(best_top1.detach().cpu().numpy())}", file=logfile, flush=True)
    # Let us evaluate by default
    if args.rank == 0:
        logfile_val = open(os.path.join(log_folder, 'val_by_hl.txt'), 'w')
        val_acc, _, _, _ = evaluate(val_loader, model, val_criterion, gpu_id=args.gpu, rank=args.rank, logfile = logfile_val)
        logfile_val.close()
        logfile_train = open(os.path.join(log_folder, 'train_by_hl.txt'), 'w')
        train_acc, _, _, _ = evaluate(train_loader, model, val_criterion, gpu_id=args.gpu, rank=args.rank, logfile = logfile_train)
        logfile_train.close()
        print('Validation Acc: %.4f'%(val_acc))
    if args.rank == 0:
        logfile.close()
        summary_writer.close()


if __name__ == '__main__':
    main()
