import torch
import torch.nn.functional as F
from torchvision.datasets import ImageFolder
from utils.randaugment import RandAugment
import numpy as np
from PIL import Image
from copy import deepcopy

class fixMatchDataset(ImageFolder):
    def __init__(self, root, transform = None, target_transform = None, loader = None, if_strong = False):
        super().__init__(root = root, transform = transform, loader = loader)
        self.loader = loader
        # self.data = [loader(path) for path, _ in self.samples]
        if if_strong:
            self.strong_transform = deepcopy(transform)
            self.strong_transform.transforms.insert(0, RandAugment(3,5))
        else:
            self.strong_transform = None
    def regenerate_data(self):
        self.data = [self.loader(path) for path, _ in self.samples]
    def __getitem__(self, idx):
        img = self.data[idx]
        target = self.targets[idx]
        if isinstance(img, np.ndarray):
            img = Image.fromarray(img)
        img_w = self.transform(img) # generate weak transform
        if self.strong_transform is not None:        
            img_s = self.strong_transform(img)
            return img_w, img_s, target
        else:
            return img_w, target
        

class Get_Scalar:
    def __init__(self, value):
        self.value = value
        
    def get_value(self, iter):
        return self.value
    
    def __call__(self, iter):
        return self.value

def ce_loss(logits, targets, use_hard_labels=True, reduction='none'):
    """
    wrapper for cross entropy loss in pytorch.
    
    Args
        logits: logit values, shape=[Batch size, # of classes]
        targets: integer or vector, shape=[Batch size] or [Batch size, # of classes]
        use_hard_labels: If True, targets have [Batch size] shape with int values. If False, the target is vector (default True)
    """
    if use_hard_labels:
        return F.cross_entropy(logits, targets, reduction=reduction)
    else:
        assert logits.shape == targets.shape
        log_pred = F.log_softmax(logits, dim=-1)
        nll_loss = torch.sum(-targets*log_pred, dim=1)
        return nll_loss

def consistency_loss(logits_w, logits_s, name='ce', T=1.0, p_cutoff=0.0, use_hard_labels=True):
    assert name in ['ce', 'L2']
    logits_w = logits_w.detach()
    if name == 'L2':
        assert logits_w.size() == logits_s.size()
        return F.mse_loss(logits_s, logits_w, reduction='mean')
    
    elif name == 'L2_mask':
        pass

    elif name == 'ce':
        pseudo_label = torch.softmax(logits_w, dim=-1)
        max_probs, max_idx = torch.max(pseudo_label, dim=-1)
        mask = max_probs.ge(p_cutoff).float()
        
        if use_hard_labels:
            masked_loss = ce_loss(logits_s, max_idx, use_hard_labels, reduction='none') * mask
        else:
            pseudo_label = torch.softmax(logits_w/T, dim=-1)
            masked_loss = ce_loss(logits_s, pseudo_label, use_hard_labels) * mask
        return masked_loss.mean(), mask.mean()

    else:
        assert Exception('Not Implemented consistency_loss')


